##################################################################################################################################
#                                                   BinEGEE                                                                      #
##################################################################################################################################
##           Author: Shuai Wang tutuwang@bu.edu                                                                                  #
##          Version: beta (intended for family data only)                                                                        #
##  Date of Release: 02/28/2015                                                                                                  #
##################################################################################################################################
##  Description of the arguments                                                                                                ##
##  formula   : the formula of the binary phenotype without the genotype term. e.g. bmi~age                                     ##
##  g         : the column name of the genotype in the dataframe data, must be quoted in quotation marks, e.g. "g1"             ##
##  id        : the column name of the id identifier in the dataframe data, must be quoted in quotation marks, e.g. "ID"        ##
##  famid     : the column name of the famid identifier in the dataframe data, must be quoted in quotation marks, e.g. "FAMID"  ##
##  data      : the name of the dataframe.                                                                                      ##
##  fullkins  : the kinship matrix                                                                                              ##
##  tol       : the convergence criteria                                                                                        ##
##  restricted: by default restricted=F, the unrestricted estimation is implemented                                             ##
##################################################################################################################################


BinEGEE=function(formula, g, id, famid, data, fullkins, tol=10^(-8), restricted=FALSE ){
fullkins=as.matrix(fullkins)
library(gdata)
phen=paste(formula)[2]
cov=paste(formula)[3]
if ((grepl("-1", deparse(formula))==T)  |(grepl("- 1", deparse(formula))==T)){
   
    cov=sapply(strsplit(cov,"+", fixed=T), function(x){trim(unlist(strsplit(x, "-", fixed=T)))})
    cov=cov1[cov!="1"]
} else {
    cov=unlist(strsplit(cov, "+", fixed=T))
    cov=sapply(cov, trim)
}

cov=cov[nchar(cov)>0]

data=data[, unique(c(id, famid, phen, cov, g))]
data=data[apply(data, 1, function(x){all(!is.na(x))}),]
kmat=as.matrix(fullkins)
X_ALL=cbind(rep(1, nrow(data)), data[, cov])
beta<-as.vector(coef(glm(formula, data=data, family="binomial")))
beta=matrix(beta, ncol=1)

thai=1
sqerror=1
iter=0

while (sqerror>=tol){
  
  Ustar=matrix(0, ncol=(ncol(X_ALL) + 1), nrow=(ncol(X_ALL) + 1 ))
  U=matrix(0, ncol=1, nrow=(ncol(X_ALL) + 1))
 
  

  for (i in  1:N) {
    
 
    v.famid <- data$famid==i
    ni=sum(v.famid)
    id.tmp=as.character(data[v.famid, id])
  
    Sigmakin_i = kmat[id.tmp, id.tmp]
       
       X=X_ALL[v.famid,]

    if (ncol(X_ALL)==1){
      X=matrix(X, ncol=1)
    } else{
      X=as.matrix(X)
    }
     
    pred=X%*%beta
    exp.pred = exp(pred)
    mu=exp.pred/(1+exp.pred)
       phen_i=data[v.famid, phen]
    
    si = as.vector(matrix(c(phen_i-mu), ncol=1) %*% matrix(c(phen_i-mu), nrow=1))
    D = as.numeric(exp.pred/(1+exp.pred)^2) * X  
    
   R.w=2*Sigmakin_i*thai
   diag(R.w)=1
 
   Vi=diag(sqrt(as.vector(mu*(1-mu))))%*%R.w%*%diag(sqrt(as.vector(mu*(1-mu))))

         
  
   solve.Vi = solve(Vi)
    sigma_i = as.vector(Vi)

      
    R.w_thai=2*Sigmakin_i
    diag(R.w_thai)=0
    Var_bin_thai=diag(sqrt(as.vector(mu*(1-mu))))%*%R.w_thai%*%diag(sqrt(as.vector(mu*(1-mu))))
Fi=as.vector(-solve.Vi %*% Var_bin_thai %*% solve.Vi)
    
 
    mu.vec=as.vector(mu)
    Vi_alpha=as.vector(Var_bin_thai)
    Vi_beta= sapply(1:ncol(X_ALL), function(x){
Vi_bin_beta=diag(1/2/sqrt(mu.vec*(1-mu.vec))*(1-2*mu.vec)*(1-mu.vec)*mu.vec*X[,x])%*%R.w%*%diag(sqrt(mu.vec*(1-mu.vec)))
+diag(sqrt(mu.vec*(1-mu.vec)))%*%R.w%*%diag(1/2/sqrt(mu.vec*(1-mu.vec))*(1-2*mu.vec)*(1-mu.vec)*mu.vec*X[,x]);

return(as.vector(Vi_bin_beta))
  })

      
           
          
    
  
    
    Utemp <- t(D) %*% solve.Vi
    Ustar_entry = cbind(rbind(Utemp %*% D, t(Fi) %*% Vi_beta), 
                        rbind(matrix( 0, nrow = ncol(X_ALL), ncol=1),t(Fi) %*% Vi_alpha) )
    U_entry =rbind(Utemp %*% c(phen_i-mu), t(Fi) %*% (si-sigma_i))
    
    #katia: in this case we do not need this
    #list(Ustar_entry, U_entry)
    Ustar <- Ustar + Ustar_entry
    U <- U + U_entry
  } 

  
  par.old=matrix(c(beta,  thai) , ncol=1)
  par.new=par.old+solve(Ustar) %*% U        
  sqerror=sum((par.old-par.new)^2)
  beta=par.new[1:ncol(X_ALL)]
  beta=matrix(beta, ncol=1)
    if (par.new[(ncol(X_ALL)+1)]>=0){
    thai=par.new[(ncol(X_ALL)+1)]
  }  
 
  iter=iter+1
 
}    


X_ALL=cbind(X_ALL, data[,g])
#katia: again: for some reason the dimentions are wrong - fix it



beta=c(beta,0)
sandwich_I=matrix(0, ncol=(ncol(X_ALL)), nrow=(ncol(X_ALL)))
U=matrix(0, ncol=1, nrow=(ncol(X_ALL)))
Ustar=matrix(0, ncol=(ncol(X_ALL)), nrow=(ncol(X_ALL)))


for(i in 1:N){
     v.famid <- data$famid==i
    ni=sum(v.famid)
    id.tmp=as.character(data[v.famid, id])
  
    Sigmakin_i = kmat[id.tmp, id.tmp]
        
    X=X_ALL[v.famid,]


    if (ncol(X_ALL)==1){
      X=matrix(X, ncol=1)
    } else{
       X=as.matrix(X)
    }
    
       
 
     pred=X%*%beta
     exp.pred = exp(pred)
    mu=exp.pred/(1+exp.pred)
     phen_i=data[v.famid, phen]
    
    si = as.vector(matrix(c(phen_i-mu), ncol=1) %*% matrix(c(phen_i-mu), nrow=1))
    
 
    D = as.numeric(exp.pred/(1+exp.pred)^2) * X
    
   R.w=2*Sigmakin_i*thai
   diag(R.w)=1
 
   Vi=diag(sqrt(as.vector(mu*(1-mu))))%*%R.w%*%diag(sqrt(as.vector(mu*(1-mu))))

         
  
   solve.Vi = solve(Vi)
    sigma_i = as.vector(Vi)
sandwich_I=sandwich_I+t(D)%*%solve.Vi%*%matrix(c( phen_i-mu), ncol=1) %*% matrix(c( phen_i-mu), nrow=1)%*%solve.Vi%*%D
         Ustar.fac=t(D)%*%solve.Vi

    U.entry=Ustar.fac%*%matrix(c(phen_i-mu), ncol=1)
    U=U+U.entry
   Ustar=Ustar+Ustar.fac%*%D


} 


U_s=U[ncol(X_ALL)]
I_s=sandwich_I[ncol(X_ALL), ncol(X_ALL)]
score_stats=U_s^2/I_s


p.wang=pchisq(score_stats, df=1, lower.tail=F)
C=cbind(-Ustar[ncol(X_ALL), (1:(ncol(X_ALL)-1))]%*%solve(Ustar[(1:(ncol(X_ALL)-1)), (1:(ncol(X_ALL)-1))]),1)
robust_stats=(C%*%U)^2/(C%*%sandwich_I%*%t(C))
p.score=pchisq(robust_stats, df=1, lower.tail=F)



if (restricted==FALSE){
    res.out=list(beta=beta,  thai=thai,  iter=iter, statistic=robust_stats, p=p.score)
} else if (restricted==TRUE){
    res.out=list(beta1=beta, thai=thai,  iter=iter, statistic=score_stats, p=p.wang)
} else {
    stop ("please provide either a TRUE or FALSE value for the argument restricted!")        
}      
return(res.out)
}




